
/* 
 * Copyright (c) 2003, Tetsuhiko Koto.
 * Licensed under the GPL.
 * 
 */

import java.util.*;
import java.util.regex.*;
import java.io.*;
import org.apache.tools.ant.*;
import org.apache.tools.ant.types.*;
import org.apache.tools.ant.taskdefs.*;
import org.apache.tools.ant.util.*;

public class HTMLJavaDocTask2 extends Javadoc {
    private String locale = null;
    public void setLocale(String locale) {
        this.locale = locale;
    }
    
    private Path depsrc = null;
    public void setDepsrc(Path path) {
        depsrc = path;
    }
    
    private File dest = null;
    public void setDestdir(File dir) {
        dest = dir;
    }
    
    private String encoding;
    public void setEncoding(String encoding) {
        this.encoding = encoding;
    }
    
    private Path sourcePath = null;
    public Path createSourcepath() {
        if (sourcePath == null) {
            sourcePath = new Path(getProject());
        }
        return sourcePath.createPath();
    }
    
    private List fileSets = new LinkedList();
    public void addFileset(FileSet fs) {
        fileSets.add(fs);
    }
    
    private File executable = null;
    public void setExecutable(File file) {
        executable = file;
    }
    
    private Commandline cmdl = new Commandline();
    public Commandline.Argument createArg() {
        return cmdl.createArgument();
    }
    
    private List replacements = null;
    public Replacement createReplacement() {
        if (replacements == null)
            replacements = new LinkedList();
        Replacement result = new Replacement();
        replacements.add(result);
        return result;
    }
    public static class Replacement {
        private Pattern old;
        public void setOld(String old) {
            this.old = Pattern.compile(old);
        }
        private String hot;
        public void setNew(String hot) {
            this.hot = hot;
        }
        
        public String toString() {
            return old + " => " + hot;
        }
    }
    
    public void execute() throws BuildException {
        if (depsrc == null)
            throw new BuildException("no depsrc");
        if (dest == null)
            throw new BuildException("no destdir");
        
        String[] list = depsrc.list();
        for (int j = 0; j < list.length; j++) {
            File base = getProject().resolveFile(list[j]);
            
            DirectoryScanner scanner = new DirectoryScanner();
            scanner.setBasedir(base);
            scanner.setIncludes(new String[] { "**" });
            scanner.scan();
            String[] files = scanner.getIncludedFiles();
            
            MergingMapper mapper = new MergingMapper();
            mapper.setTo(new File(dest, "index.html").toString());
            
            files = new SourceFileScanner(this).restrict(files, base, dest, mapper);
            if (files.length > 0) {
                try {
                    execute2();
                    if (replacements != null) {
                        boolean[] used = new boolean[replacements.size()];
                        postprocess(used, dest);
                        for (int k = 0; k < used.length; k++) {
                            if (!used[k])
                                System.out.println("Warning: not used " + replacements.get(k));
                        }
                    }
                } catch (IOException x) {
                    throw new BuildException(x);
                }
                return;
            }
        }
    }
    
    private FileUtils fileUtils = FileUtils.newFileUtils();
    private File tmpFile = null;
    private void execute2() throws BuildException, IOException {
        List arguments = new LinkedList();
        
        if (locale != null) {
            arguments.add("-locale");
            arguments.add(locale);
        }
        
        arguments.addAll(Arrays.asList(cmdl.getArguments()));
        
        arguments.add("-d");
        arguments.add(dest.getCanonicalPath());
        
        if (encoding != null) {
            arguments.add("-encoding");
            arguments.add(encoding);
            arguments.add("-charset");
            arguments.add(encoding);
            arguments.add("-docencoding");
            arguments.add(encoding);
        }
        
        if (sourcePath != null) {
            arguments.add("-sourcepath");
            arguments.add(sourcePath.toString());
        }
        
        Iterator it1 = fileSets.iterator();
        while (it1.hasNext()) {
            FileSet fileSet = (FileSet) it1.next();
            File baseDir = fileSet.getDir(getProject());
            DirectoryScanner scanner = fileSet.getDirectoryScanner(getProject());
            String[] files = scanner.getIncludedFiles();
            for (int i = 0; i < files.length; i++) {
                File file = new File(baseDir, files[i]);
                arguments.add(file.getCanonicalPath());
            }
        }
        
        if (tmpFile == null) {
            tmpFile = fileUtils.createTempFile("javadoc", "", null);
            tmpFile.deleteOnExit();
            //toExecute.createArgument().setValue("@" + tmpList.getAbsolutePath());
        }
        
        PrintWriter writer = new PrintWriter(new FileWriter(tmpFile));
        Iterator it0 = arguments.iterator();
        while (it0.hasNext()) {
            String arg = (String) it0.next();
            if (arg.indexOf(' ') >= 0)
                arg = '"' + arg + '"';
            writer.println(arg);
        }
        if (writer.checkError())
            throw new IOException("Cannot write to " + tmpFile);
        writer.close();
        
        final ExecTask cmd = (ExecTask) getProject().createTask("exec");
        if (executable != null) {
            cmd.setExecutable(executable.getCanonicalPath().toString());
        } else {
            cmd.setExecutable(JavaEnvUtils.getJdkExecutable("javadoc"));
        }
        cmd.createArg().setValue("@" + tmpFile.getCanonicalPath());
        cmd.setFailonerror(true);
        cmd.setTaskName(getTaskName());
        cmd.execute();
    }
    
    private void postprocess(boolean[] used, File file) throws IOException {
        if (file.isDirectory()) {
            File[] files = file.listFiles();
            for (int i = 0; i < files.length; i++)
                postprocess(used, files[i]);
            return;
        }
        
        if (file.getName().endsWith(".html"))
            replace(used, file);
    }
    
    private void replace(boolean[] used, File file) throws IOException {
        StringBuffer master = new StringBuffer((int) Math.max(10, Math.min(Integer.MAX_VALUE, file.length())));
        Reader reader = new InputStreamReader(new FileInputStream(file), encoding);
        char[] buffer = new char[1024];
        for(;;) {
            int size = reader.read(buffer);
            if (size < 0)
                break;
            master.append(buffer, 0, size);
        }
        reader.close();
        
        Replacement[] replacements = (Replacement[]) this.replacements.toArray(new Replacement[0]);
        Matcher[] matchers = new Matcher[replacements.length];
        int[] indexes = new int[replacements.length];
        for (int i = 0; i < replacements.length; i++) {
            matchers[i] = replacements[i].old.matcher(master);
            indexes[i] = -1;
        }
        
        StringBuffer result = new StringBuffer(master.length() * 2);
        int index = 0;
    loop:
        for (;;) {
            int min = master.length();
            for (int j = 0; j < indexes.length; j++) {
                if (indexes[j] < index) {
                    if (matchers[j].find(index)) {
                        indexes[j] = matchers[j].start();
                        //System.out.println("**match** " + matchers[j].pattern().pattern() + "  //  " + master.subSequence(matchers[j].start(), matchers[j].end()));
                    } else {
                        indexes[j] = master.length();
                    }
                }
                min = Math.min(min, indexes[j]);
            }
            
            //System.out.println("**skip** " + master.subSequence(index, min));
            result.append(master.subSequence(index, min));
            index = min;
            if (index == master.length())
                break;
            
            for (int j = 0; j < indexes.length; j++) {
                if (indexes[j] == index) {
                    used[j] = true;
                    //System.out.println("**old** " + master.subSequence(matchers[j].start(), matchers[j].end()));
                    StringBuffer replacement = new StringBuffer();
                    Replacement rep = replacements[j];
                    for (int k = 0; k < rep.hot.length(); k++) {
                        char ch = rep.hot.charAt(k);
                        if (ch != '\\' || k == rep.hot.length() - 1) {
                            replacement.append(ch);
                        } else {
                            char next = rep.hot.charAt(++k);
                            if (next == ch) {
                                replacement.append(ch);
                            } else {
                                int group = Character.digit(next, Character.MAX_RADIX);
                                if (group == -1) {
                                    replacement.append(ch);
                                    k--;
                                } else {
                                    String str = matchers[j].group(group);
                                    if (str != null)
                                        replacement.append(str);
                                }
                            }
                        }
                    }
                    result.append(replacement);
                    //System.out.println("**new** " + replacement);
                    index = matchers[j].end();
                    break;
                }
            }
        }
        
        Writer writer = new OutputStreamWriter(new FileOutputStream(file), encoding);
        writer.write(result.toString());
        writer.close();
    }
    
    private static boolean startsWith(StringBuffer buffer, int index, String str) {
        if (index + str.length() > buffer.length())
            return false;
        
        return buffer.substring(index, index + str.length()).equals(str);
    }
    
}

