
package notavacc2ebnf.parser;

import java.io.*;
import java.util.*;

public class Parser extends Syntax {
    protected Node createNode(int symbolID, NodeInitializationParameters parameters) {
        switch (symbolID) {
        case InlineExpression.ID:
            return new InlineExpressionReplacement(parameters);
        }
        Node result = super.createNode(symbolID, parameters, true);
        if (result instanceof RestrictorExpression) {
            return ((RestrictorExpression) result).operand();
        } else if (result instanceof LabeledExpression) {
            return ((LabeledExpression) result).operand();
        } else if (result instanceof BinaryExpression) {
            BinaryExpression n = (BinaryExpression) result;
            
            class SequentialExpression extends Default.SequentialExpression {
                SequentialExpression(Expression op1, Expression op2) {
                    super(new LinkedList(), Arrays.asList(new Node[] { op1, op2 }));
                }
            }
            class OptionalExpression extends Default.SelectiveExpression {
                OptionalExpression(Expression op1) {
                    super(new LinkedList(), Arrays.asList(new Node[] { op1, new Default.SequentialExpression(new LinkedList(), Collections.EMPTY_LIST) }));
                }
            }
            class PlusExpression extends Default.PlusExpression {
                PlusExpression(Expression op1, Token mark) {
                    super(new LinkedList(), op1, mark);
                }
            }
            
            Default.Expression op2op1star = new OptionalExpression(new PlusExpression(new SequentialExpression(n.operand2(), n.operand1()), n.mark()));
            Default.Expression oneOrMore = new SequentialExpression(n.operand1(), op2op1star);
            
            if (result instanceof OneOrMoreSeparatedExpression) {
                return oneOrMore;
            } else {
                assert result instanceof ZeroOrMoreSeparatedExpression;
                Default.Expression zeroOrMore = new OptionalExpression(oneOrMore);
                return zeroOrMore;
            }
        }
        return result;
    }
    
    protected LexicalAnalyzer createLexicalAnalyzer(String sourceName, CharSequence text, int tabStop) throws ParseException {
        return new Default.LexicalAnalyzer(sourceName, text, tabStop) {
            protected char nextChar() throws ParseException {
                return super.nextUnicodeEscapedChar();
            }
            public Token next() throws ParseException {
                Token result = super.next();
                if (result.getSymbolID() == TOKEN_IDENTIFIER) {
                    String image = result.getImage();
                    int i = 0;
                    for (; i < image.length(); i++) {
                        char ch = image.charAt(i);
                        if (i == 0) {
                            if (!Character.isJavaIdentifierStart(ch))
                                break;
                        } else {
                            if (!Character.isJavaIdentifierPart(ch))
                                break;
                        }
                    }
                    if (i < image.length())
                        throw new ParseException("The " + i + "th letter \\u" + Integer.toHexString(image.charAt(i)) +  " of the token " + result + " cannot be used.", result);
                }
                return result;
            }
        };
    }
    
    private class InlineExpressionReplacement extends Default.InlineExpression {
        public InlineExpressionReplacement(NodeInitializationParameters parameters) {
            super(parameters, true);
            name = new SymbolNameReplacement();
        }
        
        private final SymbolName name;
        public SymbolName name() {
            return name;
        }
        
        private class SymbolNameReplacement extends Default.Node implements SymbolName {
            public SymbolNameReplacement() {
                super(Collections.EMPTY_LIST);
            }
            public Token identifier() {
                return typeDefinition().identifier();
            }
        }
    }
    
}

