
package jp.gr.java_conf.koto.notavacc.fa;

import java.util.*;
import jp.gr.java_conf.koto.notavacc.util.*;

public class XFAStates {
    public static Set getAllStates(XNFAState start) {
        return getAllStates((State) start);
    }
    public static Set getAllStates(XSimpleFAState start) {
        return getAllStates((State) start);
    }
    public static Set getAllStates(State start) {
        return getAllStates(Collections.singleton(start));
    }
    public static Set getAllStates(Set starts) {
        Set result = new LinkedHashSet();
        getAllStates(result, starts);
        return result;
    }
    private static void getAllStates(Set result, Collection states) {
        Iterator it0 = states.iterator();
        while (it0.hasNext()) {
            State state = (State) it0.next();
            if (result.add(state)) {
                getAllStates(result, state.getMapInputToNextStates().values());
                getAllStates(result, state.getEpsilonMoveStates());
            }
        }
    }
    
    public static Set find(XNFAState start, Set attributes) {
        return find(start, attributes);
    }
    public static Set find(XSimpleFAState start, Set attributes) {
        return find(start, attributes);
    }
    public static Set find(State start, Set attributes) {
        Set result = new LinkedHashSet();
        Iterator it9 = getAllStates(start).iterator();
        while (it9.hasNext()) {
            State state = (State) it9.next();
            if (state.getAttributes().equals(attributes))
                result.add(state);
        }
        return result;
    }
    
    public static Set epsilonMove(Collection states) {
        Set result = new LinkedHashSet();
        epsilonMove(result, states);
        return result;
    }
    private static void epsilonMove(Set result, Collection states) {
        Iterator it0 = states.iterator();
        while (it0.hasNext()) {
            State state = (State) it0.next();
            if (result.add(state))
                epsilonMove(result, state.getEpsilonMoveStates());
        }
    }
    
    
    public static XDFAState toDFAState(XNFAState start) {
        return toDFAState(start, new TransitionMapMerger() {
            public Map merge(Map[] mapsInputToStates) {
                MultiMap result = new MultiMap();
                for (int i = 0; i < mapsInputToStates.length; i++) {
                    Iterator it2 = mapsInputToStates[i].entrySet().iterator();
                    while (it2.hasNext()) {
                        Map.Entry entry = (Map.Entry) it2.next();
                        Object input = entry.getKey();
                        Collection nextStates = (Collection)  entry.getValue();
                        result.addAll(input, nextStates);
                    }
                }
                return result;
            }
        });
    }
    public static XDFAState toDFAState(XNFAState start, TransitionMapMerger merger) {
        return new NDConvertor(merger).create(Collections.singleton(start));
    }
    public static interface TransitionMapMerger {
        public Map merge(Map[] inputToStates);
    }
    
    private static class NDConvertor {
        private final TransitionMapMerger merger;
        public NDConvertor(TransitionMapMerger merger) {
            this.merger = merger;
        }
        
        private final Map map = new HashMap();
        public XDFAState create(Collection states) {
            XDFAState result = (XDFAState) map.get(states);
            if (result == null) {
                final Collection states2 = epsilonMove(states);
                result = (XDFAState) map.get(states2);
                if (result == null) {
                    result = new XDFAState() {
                        private Map inputToNext;
                        public Map getMapInputToNextState() {
                            if (inputToNext == null)
                                inputToNext = createMap(states2);
                            return inputToNext;
                        }
                        private Set attributes;
                        public Set getAttributes() {
                            if (attributes == null) {
                                attributes = new LinkedHashSet();
                                Iterator it4 = states2.iterator();
                                while (it4.hasNext()) {
                                    XNFAState state = (XNFAState) it4.next();
                                    attributes.addAll(state.getAttributes());
                                }
                            }
                            return attributes;
                        }
                    };
                }
                map.put(states, result);
                map.put(states2, result);
            }
            return result;
        }
        
        private Map createMap(Collection states) {
            List mapsInputToStates = new ArrayList(states.size());
            Iterator it1 = states.iterator();
            while (it1.hasNext()) {
                XNFAState state = (XNFAState) it1.next();
                mapsInputToStates.add(state.getMapInputToNextStates());
            }
            Map map = merger.merge((Map[]) mapsInputToStates.toArray(new Map[mapsInputToStates.size()]));
            
            Map result = new LinkedHashMap();
            Iterator it3 = map.entrySet().iterator();
            while (it3.hasNext()) {
                Map.Entry entry = (Map.Entry) it3.next();
                Object input = entry.getKey();
                Collection nextStates = (Collection)  entry.getValue();
                result.put(input, create(nextStates));
            }
            return result;
        }
    }
    
    
    public static XDFAState translateAttributes(XDFAState start, AttributesTranslator translator) {
        return new AttributesTranslater(translator).create(start);
    }
    public static interface AttributesTranslator {
        public Set translate(Set attributes);
    }
    
    private static class AttributesTranslater {
        private final AttributesTranslator translator;
        public AttributesTranslater(AttributesTranslator translator) {
            this.translator = translator;
        }
        
        private final Map map = new HashMap();
        private XDFAState create(final XDFAState state) {
            XDFAState result = (XDFAState) map.get(state);
            if (result == null) {
                result = new XDFAState() {
                    private Map inputToNext;
                    public Map getMapInputToNextState() {
                        if (inputToNext == null) {
                            inputToNext = new LinkedHashMap();
                            Iterator it5 = state.getMapInputToNextState().entrySet().iterator();
                            while (it5.hasNext()) {
                                Map.Entry entry = (Map.Entry) it5.next();
                                Object input = entry.getKey();
                                XDFAState next = (XDFAState) entry.getValue();
                                inputToNext.put(input, create(state));
                            }
                        }
                        return inputToNext;
                    }
                    
                    public Set getAttributes() {
                        return translator.translate(state.getAttributes());
                    }
                };
            }
            return result;
        }
    }
    
    
    public static XSimpleFAState unlock(XDFAState start) {
        Map map = new LinkedHashMap();
        Iterator it6 = getAllStates(start).iterator();
        while (it6.hasNext()) {
            XDFAState state = (XDFAState) it6.next();
            map.put(state, new XSimpleFAState());
        }
        
        Iterator it7 = map.entrySet().iterator();
        while (it7.hasNext()) {
            Map.Entry entry = (Map.Entry) it7.next();
            XDFAState state = (XDFAState) entry.getKey();
            XSimpleFAState hot = (XSimpleFAState) entry.getKey();
            
            hot.attributes.addAll(state.getAttributes());
            
            Iterator it8 = state.getMapInputToNextState().entrySet().iterator();
            while (it8.hasNext()) {
                Map.Entry entry2 = (Map.Entry) it8.next();
                Object input = entry2.getKey();
                XDFAState next = (XDFAState) entry2.getValue();
                
                hot.link(input, (XSimpleFAState) map.get(next));
            }
        }
        
        return (XSimpleFAState) map.get(start);
    }
    
    
    public static XDFAState minimalFormed(XDFAState start) {
        XDFAState deadState = new XDFAState() {
            public Map getMapInputToNextState() {
                return Collections.EMPTY_MAP;
            }
            public Set getAttributes() {
                return Collections.EMPTY_SET;
            }
        };
        
        // initialize
        Set allStates = new LinkedHashSet();
        allStates.add(deadState);
        allStates.addAll(getAllStates(start));
        List groups = new ArrayList(allStates.size());
        Map stateToGroupID = new HashMap();
        {
            Map attributesToGroupID = new HashMap();
            Iterator it = allStates.iterator();
            while (it.hasNext()) {
                XDFAState state = (XDFAState) it.next();
                
                Integer groupID = (Integer) attributesToGroupID.get(state.getAttributes());
                if (groupID == null) {
                    groupID = new Integer(groups.size());
                    groups.add(new LinkedList());
                    attributesToGroupID.put(state.getAttributes(), groupID);
                }
                
                List states = (List) groups.get(groupID.intValue());
                states.add(state);
                stateToGroupID.put(state, groupID);
            }
        }
        
        // loop
        for (;;) {
            boolean active = false;
            
            for (int index = 0; index < groups.size(); index++) {
                Collection group = (Collection) groups.get(index);
                assert !group.isEmpty();
                if (group.size() <= 1)
                    continue;
                
                MultiMap keyToStates = new MultiMap();
                Iterator it10 = group.iterator();
                while (it10.hasNext()) {
                    XDFAState state = (XDFAState) it10.next();
                    Set key = new HashSet();
                    Iterator it2 = state.getMapInputToNextState().entrySet().iterator();
                    while (it2.hasNext()) {
                        Map.Entry entry = (Map.Entry) it2.next();
                        Object input = entry.getKey();
                        XDFAState next = (XDFAState) entry.getValue();
                        key.add(Arrays.asList(new Object[] { input, stateToGroupID.get(next) }));
                    }
                    keyToStates.add(key, state);
                }
                if (keyToStates.size() <= 1)
                    continue;
                
                active = true;
                
                Iterator it = keyToStates.values().iterator();
                while (it.hasNext()) {
                    Collection states = (Collection) it.next();
                    if (it.hasNext()) {
                        Integer newID = new Integer(groups.size());
                        groups.add(states);
                        Iterator it2 = states.iterator();
                        while (it2.hasNext()) {
                            XDFAState state = (XDFAState) it2.next();
                            stateToGroupID.put(state, newID);
                        }
                    } else {
                        groups.set(index, states);
                    }
                }
                
            }
            
            if (!active)
                break;
        }
        
        
        // build result.
        
        if (groups.size() == allStates.size())
            return start;   // already minimal
        
        Integer deadIndexInteger = (Integer) stateToGroupID.get(deadState);
        int deadIndex = deadIndexInteger.intValue();
        XSimpleFAState[] resultStates = new XSimpleFAState[groups.size()];
        for (int i = 0; i < resultStates.length; i++) {
            Collection group = (Collection) groups.get(i);
            XDFAState state = (XDFAState) group.iterator().next();
            resultStates[i] = new XSimpleFAState();
            resultStates[i].attributes.addAll(state.getAttributes());
        }
        for (int i = 0; i < resultStates.length; i++) {
            if (i != deadIndex) {
                List group = (List) groups.get(i);
                XDFAState state = (XDFAState) group.iterator().next();
                XSimpleFAState resultState = resultStates[i];
                Iterator it8 = state.getMapInputToNextState().entrySet().iterator();
                while (it8.hasNext()) {
                    Map.Entry entry = (Map.Entry) it8.next();
                    Object input = entry.getKey();
                    XDFAState nextState = (XDFAState) entry.getValue();
                    Integer index = (Integer) stateToGroupID.get(nextState);
                    if (index.intValue() != deadIndex)
                        resultState.link(input, resultStates[index.intValue()]);
                }
            }
        }
        
        Integer startIndexInteger = (Integer) stateToGroupID.get(start);
        XSimpleFAState resultStart = resultStates[startIndexInteger.intValue()];
        return resultStart.toDFAState();
    }
    
}

